from typing import Iterable, Sequence, Union
import os
import pytest
import dotenv
from tests._test_helpers.node import (
    LocalNode, BaseNode, RemoteTestnetNode, LocalTestnetNode
)

from cfx_account.account import LocalAccount
from cfx_account import Account
from conflux_web3 import (
    Web3
)
from conflux_web3.types import (
    Base32Address
)

dotenv.load_dotenv()

@pytest.fixture(scope="session")
def use_testnet() -> bool:
    return bool(os.environ.get("TESTNET_SECRET")) or bool(os.environ.get("USE_TESTNET"))

@pytest.fixture(scope="session")
def worker_index(worker_id: str) -> int:
    if worker_id == "master":
        return 0
    return int(worker_id[2:])

@pytest.fixture(scope="session")
def node(use_testnet: bool, worker_index: int) -> Iterable[BaseNode]:
    if use_testnet:
        node = RemoteTestnetNode() # connection error might occur if using public RPC
        # node = LocalTestnetNode() 
        yield node
        # node.exit()
    else:
        node = LocalNode(node_name=f"sdk-test-{worker_index}", index=worker_index)
        yield node
        # node.exit()

@pytest.fixture(scope="session")
def node_url(node: BaseNode) -> str:
    return node.url

@pytest.fixture(scope="session")
def secret_key(node: LocalNode, worker_index: int, use_testnet: bool) -> Union[str, None]:
    """
    Returns:
        str: secret key with enough balance
    """
    if not use_testnet:
        return node.secrets[0]
    try:
        return node.secrets[worker_index]
    except IndexError:
        w3 = Web3(Web3.HTTPProvider(node.url))
        acct = w3.cfx.account.create()
        w3.wallet.add_account(acct)
        faucet = w3.cfx.contract(name="Faucet")
        faucet.functions.claimCfx().transact({
            "from": acct.address,
        }).executed()
        return f"0x{acct.key.hex()}"

# no scope here
@pytest.fixture
def w3(node_url: str) -> Web3:
    """
    Returns:
        Web3: a web3 instance
    """
    provider = Web3.HTTPProvider(node_url)
    w3 = Web3(provider=provider)
    return w3

@pytest.fixture(scope="session")
def account(node_url: str, secret_key: str) -> LocalAccount:
    """external_account, not supported by node
    """
    provider = Web3.HTTPProvider(node_url)
    w3 = Web3(provider=provider)
    acct = w3.account.from_key(secret_key)
    if w3.cfx.get_balance(acct.address) == 0:
        raise Exception("Unexpected exception: provided account has no balance")
    return w3.account.from_key(secret_key)

@pytest.fixture(scope="module")
def moduled_w3(node_url: str, account: LocalAccount) -> Web3:
    """
    a web3 instance for the convenience to create module shared objects
    e.g. a transaction or a contract which is required for the module
    NOTE: DON'T CHANGE PROPERTY OF THIS W3
    """
    provider = Web3.HTTPProvider(node_url)
    w3 = Web3(provider=provider)
    w3.cfx.default_account = account
    return w3

@pytest.fixture(scope="session")
def address(account: LocalAccount) -> Base32Address:
    return account.base32_address

@pytest.fixture
def embedded_accounts(w3: Web3, use_testnet: bool) -> Sequence[Base32Address]:
    if use_testnet:
        return []
    return w3.cfx.accounts
